package de.hdm_stuttgart.mi.sd1.task1;

/**
 * <p>This enum provides Body Mass Index (BMI) related methods.</p>
 *
 * <p>A person's Body Mass Index b corresponding to weight w and height h is being defined
 *    by \( b =  w / {h^2}\). We assume metric units kilogram (kg) and meter (m). The BMI's unit will thus be
 *    \( kg \over m^2\). The following diagram shows selected BMI graphs for both metric and imperial units:</p>
 *
 * <img src="doc-files/bmi.png" style="width: 70rem;" alt='6 Body mass categories'/>
 *
 * <p>The six different colors correspond to the six enum categories {@link _3_BmiCategory#UNDERWEIGHT},
 * {@link _3_BmiCategory#NORMALWEIGHT}, ..., {@link _3_BmiCategory#VERY_SEVERELY_OBESE}. </p>
 *
 */
public enum _3_BmiCategory {

    /**
     * BMI range \( [0, 18.5[\)
     */
    UNDERWEIGHT(0),
    /**
     * BMI range \( [18.5, 25[\)
     */
    NORMALWEIGHT(18.5),
    /**
     * BMI range \( [25, 30[\)
     */
    OVERWEIGHT(25),
    /**
     * BMI range \( [30, 35[\)
     */
    MODERATELY_OBESE(30),
    /**
     * BMI range \( [35, 40[\)
     */
    SEVERELY_OBESE(35),
    /**
     * BMI range \( [40, \infty[\)
     */
    VERY_SEVERELY_OBESE(40);

    final double lowerInclusiveLimit;

    _3_BmiCategory(final double lowerInclusiveLimit) {
        this.lowerInclusiveLimit = lowerInclusiveLimit;
    }

    /**
     * <p>Get a given BMI category's successor.</p>
     * <p>Example: {@link #OVERWEIGHT} is the successor category of {@link #NORMALWEIGHT}. </p>
     *
     * @return A category's successor or <code>null</code> if starting from {@link #VERY_SEVERELY_OBESE}.
     */
    public _3_BmiCategory getFollowingCategory() {

        switch (this) {
            case UNDERWEIGHT: return NORMALWEIGHT;
            case NORMALWEIGHT: return OVERWEIGHT;
            case OVERWEIGHT: return MODERATELY_OBESE;
            case MODERATELY_OBESE: return SEVERELY_OBESE;
            case SEVERELY_OBESE: return VERY_SEVERELY_OBESE;
            default: return null;
        }

// Using ordinal() instead:

//        if (ordinal() < VERY_SEVERELY_OBESE.ordinal()) {
//            return values()[ordinal() + 1];
//        } else {
//            return null;
//        }
    }

    /**
     * <p>Calculate a person's BMI category:</p>
     *
     * <p>Example: A person of 95kg and 1.80 m in height does have a BMI of
     *   \({95 / 1.8^2} \approx 29.32 \) belonging to the [25, 30[ range and thus to the
     *   <code>OVERWEIGHT</code> category.</p>
     *
     * @param weight A person's weight in kilograms
     * @param height A person's height in meters.
     * @return One of {@link #UNDERWEIGHT}, {@link #NORMALWEIGHT}, ..., {@link #VERY_SEVERELY_OBESE} categorizing
     * the person's BMI in question.
     */
    static public _3_BmiCategory getCategory(final double weight, final double height) {
        final double bmi = weight / height / height;

        // We assume categories to be ordered by BMI values. This implementation is robust
        // with respect to adding additional categories.
        for (int i = VERY_SEVERELY_OBESE.ordinal(); 0 < i; i--) {
            final _3_BmiCategory category = values()[i];
            if (category.lowerInclusiveLimit < bmi) {
                return category;
            }
        }
        return values()[0];
    }
}
