package de.hdm_stuttgart.mi.sd1.task2;

/**
 * <p>Providing zeroes (German: "Nullstellen") of quadratic polynomials.</p>
 *
 * <p>A quadratic polynomial \( p(x)  = a x² + b x + c \) is being defined by its three coefficients
 * \(a\), \(b\) and \(c\). This class limits coefficients \(a\), \(b\) and \(c\) to type <code>int</code>. A general
 * solution is being provided by:</p>
 *
 * <p>\[ x_{1,2} = {{-b \pm \sqrt{b^2 - 4 ac}} \over {2a}} \]</p>
 *
 * <p>Depending on \( b^2 - 4 ac \) being positive, zero or negative we have either two, one or no real solutions.</p>
 *
 * <p>The following sample illustrates zeroes calculation depending on given values of \(a\), \(b\) and \(c\):</p>
 *
 * <table class="goikTableDefaults">
 *     <caption>Sample code illustrating zero values calculation</caption>
 *     <tbody>
 *         <tr>
 *             <th>Code</th>
 *             <th>Result</th>
 *         </tr>
 *         <tr>
 *             <td>
 *                 <pre><code class="java"> final QuadratPolynom poly =         // Representing
 *     new QuadratPolynom(4, -3, -10); // p(x) = 4x² - 3x - 10
 *
 * double[] zeroes = poly.getZeroes();
 *
 * System.out.println("Found " + zeroes.length + " zeroes:");
 *
 * System.out.println("x_1 = " + zeroes[0]);
 * System.out.println("x_2 = " + zeroes[1]);</code></pre>
 *             </td>
 *             <td>
 *                 <pre><code class="nohighlight"> Found 2 zeroes:
 * x_1 = -1.25
 * x_2 = 2.0</code></pre>
 *             </td>
 *         </tr>
 *     </tbody>
 * </table>
 *
 * <p>We allow for re-setting a polynomial's coefficients. Continuing from the above example we have:</p>
 *
 * <table class="goikTableDefaults">
 *     <caption>Changing coefficients</caption>
 *     <tbody>
 *         <tr>
 *             <th>Code</th>
 *             <th>Result</th>
 *         </tr>
 *         <tr>
 *             <td>
 *                 <pre><code class="java"> ...
 * poly.setA(1);  // Representing
 * poly.setB(-8); // p(x) = x² -8x + 16
 * poly.setC(16);
 *
 * zeroes = poly.getZeroes();
 * System.out.println("Found " + zeroes.length + " zero:");
 *
 * System.out.println("x_1 = " + zeroes[0]);</code></pre>
 *             </td>
 *             <td>
 *                 <pre><code class="nohighlight"> ...
 * Found 1 zero:
 * x_1 = 4.0</code></pre>
 *             </td>
 *         </tr>
 *     </tbody>
 * </table>
 *
 * <p>Some polynomials do not have any (real valued) solution:</p>
 *
 * <table class="goikTableDefaults">
 *     <caption>Changing coefficients again</caption>
 *     <tbody>
 *         <tr>
 *             <th>Code</th>
 *             <th>Result</th>
 *         </tr>
 *         <tr>
 *             <td>
 *                 <pre><code class="java"> ...
 * poly.setA(1);  // Representing
 * poly.setB(0);  // p(x) = x² + 1
 * poly.setC(1);  // (No zero solution)
 *
 * zeroes = poly.getZeroes();
 * System.out.println("Found " + zeroes.length + " zero");</code></pre>
 *             </td>
 *             <td>
 *                 <pre><code class="nohighlight"> ...
 * Found 0 zero</code></pre>
 *             </td>
 *         </tr>
 *     </tbody>
 * </table>
 *
 * <p>Finally setting \(a=0\) in \(ax²\) leaves us with just a linear polynomial \(bx + c\). To
 * avoid unpredictable results the attempt shall result in throwing an
 * <a href="https://freedocs.mi.hdm-stuttgart.de/doc/openjdk-17-doc/api/java.base/java/lang/ArithmeticException.html"
 *  ><code>ArithmeticException</code></a>:</p>
 *
 * <table class="goikTableDefaults">
 *     <caption>Dealing with 0 value at square coefficient.</caption>
 *     <tbody>
 *         <tr>
 *             <th>Code</th>
 *             <th>Result</th>
 *         </tr>
 *         <tr>
 *             <td>
 *                 <pre><code class="java"> ...
 * poly.setA(0);  // Trying to represent p(x) = 1</code></pre>
 *             </td>
 *             <td>
 *                 <pre><code class="nohighlight"> ...
 * Exception in thread "main" java.lang.ArithmeticException:
 *     Square coefficient must not be zero</code></pre>
 *             </td>
 *         </tr>
 *     </tbody>
 * </table>
 *
 * <p>Likewise a <code>new QuadratPolynom(0, ..., ...)</code> constructor call setting \(a=0\) shall raise an
 * <a href="https://freedocs.mi.hdm-stuttgart.de/doc/openjdk-17-doc/api/java.base/java/lang/ArithmeticException.html">
 * <code>ArithmeticException</code></a> as well.</p>
 *
 * <section class="implementationHints">
 *    <h2 class="implementationHints">Hint:</h2>
 *
 * <p>This class is yet unimplemented. The above code snippets provide a clue to an implementation
 *    satisfying the corresponding unit tests.</p>
 *
 * </section>
 */
public class QuadratPolynom {

    private int a, b, c;

    /**
     * <p>Defining a polynomial \( p(x) = a x² + b x + c  \).</p>
     *
     * @param a Square coefficient.
     * @param b Linear coefficient.
     * @param c Constant coefficient.
     * @throws ArithmeticException  The square coefficient \( a x² \) must not be zero
     */
    public QuadratPolynom(int a, int b, int c) throws ArithmeticException {
        setA(a);
        setB(b);
        setC(c);
    }

    /**
     * <p>Re- setting the square coefficient in \( a x² \).</p>
     * @param a The desired new value.
     * @throws ArithmeticException  The square coefficient \( a x² \) must not be zero
     */
    public void setA(final int a) throws ArithmeticException {
        if (0 == a) {
            throw new ArithmeticException("Square coefficient must not be zero");
        }
        this.a = a;
    }
    /**
     * <p>Re- setting the linear coefficient in \( b x \).</p>
     * @param b The desired new value.
     */
    public void setB(final int b) {
        this.b = b;
    }
    /**
     * <p>Re- setting the constant coefficient \( c \).</p>
     * @param c The desired new value.
     */
    public void setC(final int c) {
        this.c = c;
    }

    /**
     * <p>Zeroes are all values of \(x\) solving \(a x² + b x + c = 0\).</p>
     *
     * <p>Zeros are being calculated using:</p>
     *
     * <p>\[ x = {{ -b \pm \sqrt {b² - 4 a c}} \over {2 a}} \]</p>
     *
     * <p>Modifying returned array values does not affect the called instance's internal state.</p>
     *
     * @return <p>Three distinct cases:</p>
     *
     * <dl>
     *
     *     <dt>\( b² - 4 a c &lt; 0\):</dt>
     *     <dd><p>An <code>double[0]</code> empty array indicating the non-existence of any real valued solution.</p></dd>
     *
     *     <dt>\( b² - 4 a c = 0\):</dt>
     *     <dd><p>An <code>double[1]</code> array containing the single zero {\(-b \over {2 a} \)}.</p></dd>
     *
     *     <dt>\( b² - 4 a c &gt; 0\):</dt>
     *     <dd><p>An <code>double[2]</code> array containing two distinct zeroes
     *     \( {1\over {2 a}}\left( -b - \sqrt {b² - 4 a c}\right) \)
     *     and \( {1\over {2 a}}\left( -b + \sqrt {b² - 4 a c}\right) \) ordered by size.</p></dd>
     *
     * </dl>
     *
     * <section class = "implementationHints">
     *   <h4>Implementation hints:</h4>
     *
     *   <ul>
     *       <li><a href=
     *       "https://freedocs.mi.hdm-stuttgart.de/lib/openjdk-17-doc/api/java.base/java/lang/Math.html#sqrt(double)"
     *       >Math.sqrt(...)</a></li>
     *
     *   </ul>
     *
     * </section>
     */
    public double[] getZeroes() {
        final long radicand = ((long) b) * ((long) b) - 4 * ((long) a) * ((long) c) ;
//        final int radicand = b * b - 4 * a * c; // failing at higher values due to overflow

        if (0 > radicand) {
            return new double[]{};
        } else if (0 == radicand) {
            return  new double[]{-b / 2. / a};
        } else {
            final double radicandRoot = Math.sqrt(radicand);

            if (0 < a) {
                return new double[]{(-b - radicandRoot) / 2 / a, (-b + radicandRoot) / 2 / a};
            } else {
                return new double[]{(-b + radicandRoot) / 2 / a, (-b - radicandRoot) / 2 / a};
            }

        }
    }
    private QuadratPolynom(){/* Ignore me: My sole purpose is suppressing default constructor javadoc generation */}
}
